#include "femProblem.h"
#include "femEntity.h"
#include "femElementSet.h"
#include "femGroup.h"
#include "femMedium.h"
#include "femDirichlet.h"
#include "femNeumann.h"
#include "femGmsh.h"
#include <fstream>

namespace fem
{

Problem::Problem() : quadrature("Gauss2")
{
    gmsh::initialize();
    // General.Verbosity
    // (0: silent except for fatal errors, 1: +errors, 2: +warnings, 3: +direct, 4: +information, 5: +status (default), 99: +debug)
    gmsh::option::setNumber("General.Verbosity", 2);
    gmsh::option::setNumber("General.Terminal", 1); // Should information be printed on the terminal? 0: no (default) ; 1: yes
}

Problem::~Problem()
{
    for (auto &e : entities)
        delete e;
}

/// fill internal helper objects from gmsh data

void Problem::sync()
{
    // -- entities
    std::cout << "retrieving entities...\n";
    gmsh::vectorpair dimTags;
    gmsh::model::getEntities(dimTags);
    for (auto &p : dimTags)
        entities.push_back(new Entity(p.first, p.second));
    // create useful maps
    for (auto e : entities)
        entities_by_dimtag[std::make_pair(e->dim, e->tag)] = e;

    // -- physical groups
    std::cout << "retrieving physical groups...\n";
    dimTags.clear();
    gmsh::model::getPhysicalGroups(dimTags);
    for (auto &p : dimTags)
        groups.push_back(new Group(p.first, p.second));
    // create useful maps
    for (auto g : groups)
    {
        groups_by_dimtag[std::make_pair(g->dim, g->tag)] = g;
        groups_by_name[g->name] = g;
    }

    // link entities <=> groups
    for (auto g : groups)
        g->link_to_entities(entities_by_dimtag);
    for (auto e : entities)
        e->link_to_groups(groups_by_dimtag);

    // -- nodes
    // get all the node tags and create a map ("node tag" => "index")
    // that could be used to store things at nodes in separate arrays
    std::vector<double> coord;
    std::vector<double> parametricCoord;
    gmsh::model::mesh::getNodes(nodeTags, coord, parametricCoord);
    for (size_t i = 0; i < nodeTags.size(); ++i)
        nodemap[nodeTags[i]] = i;

    //std::cout << *this;
}

void Problem::load(std::string const &filename,
                   std::map<std::string, double> const &pars,
                   int order)
{
    // std::cout << "pars=\n";
    // for (auto &p : pars)
    //     std::cout << p.first << " : " << p.second << '\n';

    std::string fext;
    auto idx = filename.rfind('.');
    if(idx != std::string::npos)
        fext = filename.substr(idx);

    if (fext!=".geo" && fext!=".msh")
        throw std::runtime_error("unknown file extension ("+fext+")");

    // load file and initialise gmsh
    // gmsh::initialize();
    for (auto &p : pars)
    {
        if(debug_level()>1)
            std::cout << "setting parameter " << p.first << "=" << p.second << '\n';
        std::vector<double> v;
        v.push_back(p.second);
        gmsh::onelab::setNumber(p.first, v);
    }
    std::cout << "loading " << filename << '\n';
    if(! std::ifstream(filename.c_str()).good() ) // not tested by gmsh
        throw std::runtime_error("file " + filename + " does not exist");
    gmsh::open(filename);

    // generate mesh
    if (fext==".geo")
    {
        gmsh::model::mesh::generate(3);
        gmsh::model::mesh::setOrder(order); // after generation
        quadrature = "Gauss" + std::to_string(2*order); // change default quadrature rule 
    }

    // do not sync now, so that we can add things such as
    // additional physical groups
}

FEM_API std::ostream &
operator<<(std::ostream &out, Problem const &obj)
{
    out << "Problem:\n";
    out << '\t' << obj.entities.size() << " entities\n";
    if(debug_level()>1)
        for (auto e : obj.entities)
        {
            out << "\t\t" << *e << '\n';
            for (auto m : e->esets)
                out << "\t\t\tmesh: " << *m << '\n';
        }
    out << '\t' << obj.groups.size() << " groups\n";
    if(debug_level()>1)
        for (auto g : obj.groups)
            out << "\t\t" << *g << '\n';
    out << '\t' << obj.media.size() << " media\n";
    if(debug_level()>1)
        for (auto m : obj.media)
            out << "\t\t" << *m << '\n';
    out << '\t' << obj.dirichlets.size() << " dirichlets\n";
    if(debug_level()>1)
        for (auto bc : obj.dirichlets)
            out << "\t\t" << *bc << '\n';
    out << '\t' << obj.neumanns.size() << " neumanns\n";
    if(debug_level()>1) // TODO: si >20 ... afficher les 10 premiers et 10 derniers   
        for (auto bc : obj.neumanns)
            out << "\t\t" << *bc << '\n';
    return out;
}

} // namespace fem
